@@ -1,8 +1,11 @@
+from __future__ import annotations
+
+import contextlib
 import csv
 import io
 import warnings
-from dataclasses import dataclass, replace
-from typing import Any, cast, Dict, Generator, List, Optional, Sequence
+from dataclasses import dataclass, field, replace
+from typing import Any, cast, Generator, Sequence
 
 import sqlalchemy
 from sqlalchemy import inspect, MetaData, Table, text
@@ -11,7 +14,7 @@ from sqlalchemy.orm import Session
 
 from databudgie.output import Console, default_console
 from databudgie.table_op import TableOp
-from databudgie.utils import join_paths, parse_table
+from databudgie.utils import join_paths, parse_table, wrap_buffer
 
 
 @dataclass
@@ -28,7 +31,7 @@ class Adapter:
     session: Session
 
     @classmethod
-    def get_adapter(cls, session: Session, dialect: Optional[str] = None) -> "Adapter":
+    def get_adapter(cls, session: Session, dialect: str | None = None) -> Adapter:
         """Determine an interface based on the dialect name from the Session (or an explicit string).
 
         Examples:
@@ -49,29 +52,35 @@ class Adapter:
             from databudgie.adapter.postgres import PostgresAdapter
 
             return PostgresAdapter(session)
-        else:
-            return cls(session)
 
-    def export_query(self, query: str, dest: io.StringIO):
+        return cls(session)
+
+    def export_query(self, query: str) -> QueryResult:
         def query_database(session: Session, query: str) -> Generator[Sequence[Any], None, None]:
             cursor = session.execute(text(query))
 
             columns: Sequence[str] = list(cursor.keys())
             yield columns
 
-            for row in cursor:
-                yield row
+            yield from cursor
 
-        writer = csv.writer(dest, quoting=csv.QUOTE_MINIMAL)
-        for row in query_database(self.session, query):
-            writer.writerow(row)
+        result = QueryResult()
+        with result.text_buffer() as text_buffer:
+            writer = csv.writer(text_buffer, quoting=csv.QUOTE_MINIMAL)
+
+            i = 0
+            for i, row in enumerate(query_database(self.session, query), start=1):
+                writer.writerow(row)
+
+        result.row_count = i
+        return result
 
     def import_csv(self, csv_file: io.TextIOBase, table: str):
         reader = csv.DictReader(csv_file, quoting=csv.QUOTE_MINIMAL)
 
-        prepared_rows: List[dict] = []
+        prepared_rows: list[dict] = []
         for row in reader:
-            new_row: Dict[str, Any] = dict(row)
+            new_row: dict[str, Any] = dict(row)
             for key, value in new_row.items():
                 if value.lower() == "true":
                     new_row[key] = True
@@ -93,6 +102,15 @@ class Adapter:
             conn.execute(table_ref.insert(), prepared_rows)
             conn.execute(text("commit"))
 
+    def execute_sql(self, sql: bytes, *, commit=False):
+        """Abstract the execution of SQL queries away from the calling code."""
+        text_sql = sql.decode("utf-8")
+        result = self.session.execute(sqlalchemy.text(text_sql))
+        if commit:
+            self.session.commit()
+
+        return result
+
     def export_schema_ddl(self, name: str):
         raise NotImplementedError()  # pragma: no cover
 
@@ -117,7 +135,7 @@ class Adapter:
         """
         raise NotImplementedError()
 
-    def collect_existing_tables(self) -> List[str]:
+    def collect_existing_tables(self) -> list[str]:
         """Find the set of all user-defined tables in a database."""
         connection = self.session.connection()
 
@@ -134,10 +152,10 @@ class Adapter:
 
         return [table.fullname for table in metadata.sorted_tables]
 
-    def collect_table_dependencies(self, table_op: TableOp, console: Console = default_console) -> List[str]:
+    def collect_table_dependencies(self, table_op: TableOp, console: Console = default_console) -> list[str]:
         raise NotImplementedError()
 
-    def collect_table_sequences(self) -> Dict[str, List[str]]:
+    def collect_table_sequences(self) -> dict[str, list[str]]:
         raise NotImplementedError()
 
     def collect_sequence_value(self, sequence_name: str) -> int:
@@ -148,13 +166,16 @@ class Adapter:
 
     def materialize_table_dependencies(
         self,
-        table_ops: List[TableOp],
+        table_ops: list[TableOp],
         reverse: bool = False,
         console: Console = default_console,
-    ) -> List[TableOp]:
+    ) -> list[TableOp]:
         tables = set()
         dependent_table_ops = []
         for table_op in table_ops:
+            if table_op.full_name is None:
+                continue
+
             tables.add(table_op.full_name)
 
             if not table_op.raw_conf.follow_foreign_keys:
@@ -163,7 +184,8 @@ class Adapter:
             dependent_tables = self.collect_table_dependencies(table_op=table_op, console=console)
             for dependent_table in dependent_tables:
                 if dependent_table not in tables:
-                    table_location = join_paths(table_op.location(), "{table}")
+                    concrete_parent_location = table_op.location().format(table=table_op.full_name)
+                    table_location = join_paths(concrete_parent_location, "{table}")
                     conf = replace(table_op.raw_conf, name=dependent_table, location=table_location)
 
                     dependent_table_op = TableOp.from_name(dependent_table, conf)
@@ -177,3 +199,19 @@ class Adapter:
             return list(reversed(dependent_table_ops)) + table_ops
 
         return table_ops + dependent_table_ops
+
+
+@dataclass
+class QueryResult:
+    buffer: io.BytesIO = field(default_factory=io.BytesIO)
+    row_count: int = 0
+
+    @contextlib.contextmanager
+    def binary_buffer(self):
+        yield self.buffer
+        self.buffer.seek(0)
+
+    @contextlib.contextmanager
+    def text_buffer(self):
+        with wrap_buffer(self.buffer) as text_buffer:
+            yield text_buffer
