@@ -1,16 +1,18 @@
+import contextlib
 import io
 import os
 import shlex
 import shutil
-import subprocess  # nosec
+import subprocess
 from typing import cast, Dict, List
 
-import psycopg2.errors
+from psycopg import Cursor, sql
 from sqlalchemy import text
 from sqlalchemy.engine import create_engine, Engine
 from sqlalchemy.engine.url import URL
+from typing_extensions import LiteralString
 
-from databudgie.adapter.base import Adapter
+from databudgie.adapter.base import Adapter, QueryResult
 from databudgie.output import Console, default_console
 from databudgie.table_op import TableOp
 
@@ -30,29 +32,45 @@ def update_url(url, database=None):
 
 
 class PostgresAdapter(Adapter):
-    def export_query(self, query: str, dest: io.StringIO):
-        engine: Engine = cast(Engine, self.session.get_bind())
-        conn = engine.raw_connection()
-        cursor: psycopg2.cursor = conn.cursor()
+    def execute_sql(self, sql: bytes, *, commit=False):
+        """Abstract the execution of SQL queries away from the calling code.
 
-        copy = "COPY ({}) TO STDOUT CSV HEADER".format(query)
-        cursor.copy_expert(copy, dest)
-        cursor.close()
-        conn.close()
+        Additionally performs pg_dump-specific SQL "cleaning" to remove statements
+        which could/would cause issues or otherwise bloat the resultant query.
+        """
+        cleaned_sql = clean_sql(sql)
+        return super().execute_sql(cleaned_sql, commit=commit)
+
+    def export_query(self, query: str) -> QueryResult:
+        engine: Engine = cast(Engine, self.session.get_bind())
+
+        result = QueryResult()
+        with result.binary_buffer() as buffer:
+            with contextlib.closing(engine.raw_connection()) as conn:
+                with cast(Cursor, conn.cursor()) as cursor:
+                    statement = sql.SQL(cast(LiteralString, f"COPY ({query}) TO STDOUT CSV HEADER"))
+
+                    with cursor.copy(statement) as copy:
+                        while data := copy.read():
+                            buffer.write(data)
+                    result.row_count = cursor.rowcount
+
+        return result
 
     def import_csv(self, csv_file: io.TextIOBase, table: str):
         engine: Engine = cast(Engine, self.session.get_bind())
-        conn = engine.raw_connection()
-        cursor: psycopg2.cursor = conn.cursor()
 
         # Reading the header line from the buffer removes it for the ingest
-        columns: List[str] = [f'"{c}"' for c in csv_file.readline().strip().split(",")]
+        columns = [sql.Identifier(c) for c in csv_file.readline().strip().split(",")]
+        statement = sql.SQL("COPY {table} ({columns}) FROM STDIN CSV").format(
+            table=sql.Identifier(*table.split(".")), columns=sql.SQL(", ").join(columns)
+        )
 
-        copy = "COPY {table} ({columns}) FROM STDIN CSV".format(table=table, columns=",".join(columns))
-        cursor.copy_expert(copy, csv_file)
-        cursor.close()
-        conn.commit()
-        conn.close()
+        with contextlib.closing(engine.raw_connection()) as conn:
+            with cast(Cursor, conn.cursor()) as cursor:
+                with cursor.copy(statement) as copy:
+                    copy.write(csv_file.read())
+                conn.commit()
 
     def export_schema_ddl(self, name: str, console: Console = default_console) -> bytes:
         if not shutil.which("pg_dump"):
@@ -61,10 +79,10 @@ class PostgresAdapter(Adapter):
 
         url = self.session.connection().engine.url
         result = pg_dump(url, f"--schema-only --schema={name} --exclude-table={name}.*")
-        result = result.replace(
-            f"CREATE SCHEMA {name};".encode("utf-8"), f"CREATE SCHEMA IF NOT EXISTS {name};".encode("utf-8")
+        return result.replace(
+            f"CREATE SCHEMA {name};".encode(),
+            f"CREATE SCHEMA IF NOT EXISTS {name};".encode(),
         )
-        return result
 
     def export_table_ddl(self, table_name: str, console: Console = default_console):
         if not shutil.which("pg_dump"):
@@ -196,7 +214,8 @@ class PostgresAdapter(Adapter):
         )
 
         results = self.session.execute(
-            collect_tables, params=dict(schema=table_op.schema, table_name=table_op.table_name)
+            collect_tables,
+            params={"schema": table_op.schema, "table_name": table_op.table_name},
         )
 
         return [row[0] for row in results]
@@ -226,13 +245,21 @@ class PostgresAdapter(Adapter):
         return result
 
     def collect_sequence_value(self, sequence_name: str) -> int:
-        return cast(int, self.session.execute(text(f"SELECT last_value from {sequence_name}")).scalar())  # nosec
+        return cast(
+            int,
+            self.session.execute(
+                text(f"SELECT last_value from {sequence_name}")  # noqa: S608
+            ).scalar(),
+        )
 
     def restore_sequence_value(self, sequence_name: str, value: int) -> int:
-        return cast(int, self.session.execute(text(f"SELECT setval('{sequence_name}', {value})")).scalar())
+        return cast(
+            int,
+            self.session.execute(text(f"SELECT setval('{sequence_name}', {value})")).scalar(),
+        )
 
 
-def pg_dump(url: URL, rest: str = "", no_comments=True) -> bytes:
+def pg_dump(url: URL, rest: str = "", no_comments=True, clean=True) -> bytes:
     parts = [f"pg_dump -h {url.host} -p {url.port} -U {url.username} -d {url.database} --no-password"]
 
     if no_comments:
@@ -245,10 +272,27 @@ def pg_dump(url: URL, rest: str = "", no_comments=True) -> bytes:
     command = shlex.split(raw_command)
 
     try:
-        result = subprocess.run(  # nosec
-            command, capture_output=True, env={**os.environ, "PGPASSWORD": str(url.password or "")}, check=True
+        result = subprocess.run(  # noqa: S603
+            command,
+            capture_output=True,
+            env={**os.environ, "PGPASSWORD": str(url.password or "")},
+            check=True,
         )
     except subprocess.CalledProcessError as e:
         raise RuntimeError(e.stderr)
 
-    return result.stdout
+    output = result.stdout
+
+    return clean_sql(output)
+
+
+def clean_sql(sql: bytes) -> bytes:
+    return b"\n".join(
+        line
+        for line in sql.splitlines()
+        if not line.startswith(b"--")
+        and not line.startswith(b"SET")
+        and not line.startswith(b"SELECT pg_catalog")
+        and not line.startswith(b"\\")
+        and line
+    )
