@@ -2,8 +2,8 @@ from __future__ import annotations
 
 import abc
 import typing
-from dataclasses import asdict, dataclass, field
-from typing import Any, Optional, Union
+from dataclasses import dataclass, field, fields
+from typing import Any
 
 from databudgie.utils import join_paths
 
@@ -17,7 +17,7 @@ class ConfigError(ValueError):
 
 class ConfigStack:
     def __init__(self, *configs: dict):
-        self.configs: typing.Tuple[dict, ...] = configs
+        self.configs: tuple[dict, ...] = configs
 
     def __getitem__(self, key):
         for config in self.configs:
@@ -43,7 +43,20 @@ class ConfigStack:
 
 class Config(metaclass=abc.ABCMeta):
     def to_dict(self) -> dict:
-        return asdict(self)
+        result = {}
+        for f in fields(self):  # type: ignore
+            v = getattr(self, f.name)
+            if isinstance(v, Config):
+                value: Any = v.to_dict()
+            elif isinstance(v, list):
+                value = [v.to_dict() if isinstance(v, Config) else v for v in v]
+            elif isinstance(v, dict):
+                value = {k: v.to_dict() if isinstance(v, Config) else v for k, v in v.items()}
+            else:
+                value = v
+
+            result[f.name] = value
+        return result
 
 
 def from_partial(cls: typing.Callable[..., F], **kwargs) -> F:
@@ -64,25 +77,32 @@ def from_partial(cls: typing.Callable[..., F], **kwargs) -> F:
 @dataclass
 class DDLConfig(Config):
     enabled: bool = False
-    location: str = "ddl"
+    location: str = "backups/ddl"
+    filename: str = "{timestamp}.{ext}"
     clean: bool = False
     strategy: str = "use_latest_filename"
 
     @classmethod
-    def from_dict(cls, ddl_config: Union[dict, bool], root_location: Optional[str] = None):
-        if isinstance(ddl_config, bool):
-            expanded_ddl_config: dict[str, Any] = {"enabled": ddl_config}
+    def from_dict(cls, ddl_config: dict | bool | str, root_location: str | None = None):
+        if isinstance(ddl_config, (bool, str)):
+            expanded_ddl_config: dict[str, Any] = {"enabled": bool(ddl_config)}
         else:
             expanded_ddl_config = ddl_config
 
-        location = join_paths(root_location, expanded_ddl_config.get("location"))
+        location = compose_root_location(root_location, expanded_ddl_config.get("location"), default="backups/ddl")
 
         # Splat into a new dict so we can override `location` without mutating
         # the original input (which may be re-read later in config parsing)
         final_ddl_config = {**expanded_ddl_config, "location": location}
 
+        if "clean" in final_ddl_config:
+            final_ddl_config["clean"] = bool(final_ddl_config["clean"])
+
         return from_partial(cls, **final_ddl_config)
 
+    def full_path(self):
+        return join_paths(self.location, self.filename)
+
 
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 # Core configuration models
@@ -91,8 +111,8 @@ class DDLConfig(Config):
 
 @dataclass
 class RootConfig(Config):
-    backup: Optional[BackupConfig] = None
-    restore: Optional[RestoreConfig] = None
+    backup: BackupConfig
+    restore: RestoreConfig
 
     @classmethod
     def from_dict(cls, raw_config: dict):
@@ -108,25 +128,19 @@ class RootConfig(Config):
         restore = RestoreConfig.from_stack(stack.push(restore_config))
         return cls(backup=backup, restore=restore)
 
-    def to_dict(self) -> dict:
-        return {
-            "backup": self.backup.to_dict() if self.backup else None,
-            "restore": self.restore.to_dict() if self.restore else None,
-        }
-
 
 @dataclass
 class TableParentConfig(typing.Generic[T], Config):
-    url: typing.Union[str, dict]
-    tables: typing.List[T]
-
+    tables: list[T]
+    connections: dict[str, Connection | str]
     ddl: DDLConfig
-    manifest: Optional[str] = None
 
-    s3: Optional[S3Config] = None
-    sentry: Optional[SentryConfig] = None
-    root_location: Optional[str] = None
-    adapter: Optional[str] = None
+    connection: Connection | str = "default"
+    manifest: str | None = None
+
+    s3: S3Config | None = None
+    root_location: str | None = None
+    adapter: str | None = None
 
     @classmethod
     @abc.abstractmethod
@@ -135,7 +149,7 @@ class TableParentConfig(typing.Generic[T], Config):
 
     @classmethod
     def from_stack(cls, stack: ConfigStack):
-        url: str = stack.get("url")
+        connection = Connection.from_raw(stack.get("url") or stack.get("connection"), name="default") or "default"
         root_location = stack.get("root_location")
 
         tables_config: list = normalize_table_config(stack.get("tables", []))
@@ -143,46 +157,93 @@ class TableParentConfig(typing.Generic[T], Config):
         tables = [table_class.from_stack(stack.push(tbl_conf), root_location) for tbl_conf in tables_config]
 
         # manifest defauls to None
-        manifest: Optional[str] = stack.get("manifest")
+        manifest: str | None = stack.get("manifest")
 
         ddl = DDLConfig.from_dict(stack.get("ddl", {}), root_location)
 
         # Optional integration configs
         s3 = S3Config.from_dict(stack.get("s3"))
-        sentry = SentryConfig.from_dict(stack.get("sentry"))
 
         adapter = stack.get("adapter")
 
+        connections = Connection.from_collection(stack.get("connections"))
+
         return cls(
-            url=url,
+            connection=connection,
             tables=tables,
             manifest=manifest,
             s3=s3,
-            sentry=sentry,
             ddl=ddl,
             root_location=root_location,
             adapter=adapter,
+            connections=connections,
         )
 
-    def to_dict(self) -> dict:
-        return {
-            "url": self.url,
-            "tables": [table.to_dict() for table in self.tables],
-            "manifest": self.manifest,
-            "ddl": self.ddl.to_dict(),
-            "s3": self.s3.to_dict() if self.s3 else None,
-            "sentry": self.sentry.to_dict() if self.sentry else None,
-            "root_location": self.root_location,
-            "adapter": self.adapter,
-        }
-
 
 @dataclass
-class BackupTableConfig(Config):
+class Connection(Config):
     name: str
+    url: str | dict
+
+    @classmethod
+    def from_raw(cls, raw: str | dict | None, *, name: str | None = None) -> Connection | str | None:
+        if raw is None:
+            return None
+
+        if isinstance(raw, str):
+            if "://" not in raw:
+                return raw
+
+            if name is None:
+                raise ConfigError(f"Connection '{raw}' requires a name")
+            return cls(name="default", url=raw)
+
+        if name is None:
+            raise ConfigError(f"Connection '{raw}' requires a name")
+
+        if "url" in raw:
+            url = raw["url"]
+        else:
+            url = {k: v for k, v in raw.items() if k != "name"}
+
+        return cls(name=name or "default", url=url)
+
+    @classmethod
+    def from_collection(cls, collection: list | dict | None) -> dict[str, Connection | str]:
+        if collection is None:
+            return {}
+
+        if isinstance(collection, list):
+            connections = []
+            names = set()
+            for c in collection:
+                connection = Connection.from_raw(c, name=c.get("name"))
+                if not isinstance(connection, Connection):
+                    raise ConfigError(
+                        "Connections must be a database connection string or a mapping of individual connection fields."
+                    )
+
+                if connection.name in names:
+                    raise ConfigError(f"Detected more than one connection with the same name: {connection.name}")
+                names.add(connection.name)
+                connections.append(connection)
+
+            return {c.name: c for c in connections}
+
+        result = {}
+        for k, c in collection.items():
+            connection = Connection.from_raw(c, name=k)
+            if connection is not None:
+                result[k] = connection
+        return result
+
+
+@dataclass
+class TableConfig(Config):
+    name: str | None = None
     location: str = "backups/{table}"
-    query: str = "select * from {table}"
-    compression: Optional[str] = None
+    filename: str = "{timestamp}.{ext}"
+    compression: str | None = None
     exclude: list = field(default_factory=list)
     ddl: bool = True
     sequences: bool = True
@@ -191,25 +252,42 @@ class BackupTableConfig(Config):
     strict: bool = False
 
     @classmethod
-    def from_stack(cls, stack: ConfigStack, root_location: Optional[str] = None):
+    def collect_values(cls, stack: ConfigStack, root_location: str | None = None):
         ddl = stack.get("ddl", True)
         if isinstance(ddl, dict):
             ddl = ddl["enabled"]
 
-        location = join_paths(root_location, stack.get("location")) or None
+        location = compose_root_location(root_location, stack.get("location"), default=cls.location)
+        filename = stack.get("filename", cls.filename)
+
+        return {
+            "name": stack.get("name"),
+            "location": location,
+            "filename": filename,
+            "compression": stack.get("compression"),
+            "exclude": stack.get("exclude"),
+            "sequences": bool(stack.get("sequences", True)),
+            "data": bool(stack.get("data", True)),
+            "ddl": bool(ddl),
+            "follow_foreign_keys": bool(stack.get("follow_foreign_keys", False)),
+            "strict": bool(stack.get("strict", False)),
+        }
+
+
+@dataclass
+class BackupTableConfig(TableConfig):
+    query: str = "select * from {table}"
+    skip_if_exists: bool = False
+
+    @classmethod
+    def from_stack(cls, stack: ConfigStack, root_location: str | None = None):
+        values = cls.collect_values(stack, root_location)
 
         return from_partial(
             cls,
-            name=stack.get("name"),
-            location=location,
+            **values,
             query=stack.get("query"),
-            compression=stack.get("compression"),
-            exclude=stack.get("exclude"),
-            sequences=stack.get("sequences", True),
-            data=stack.get("data", True),
-            ddl=stack.get("ddl", True),
-            follow_foreign_keys=stack.get("follow_foreign_keys", False),
-            strict=stack.get("strict", False),
+            skip_if_exists=bool(stack.get("skip_if_exists", False)),
         )
 
 
@@ -221,40 +299,19 @@ class BackupConfig(TableParentConfig[BackupTableConfig]):
 
 
 @dataclass
-class RestoreTableConfig(Config):
-    name: str
-    location: str = "backups/{table}"
+class RestoreTableConfig(TableConfig):
     strategy: str = "use_latest_filename"
-    compression: Optional[str] = None
-    exclude: list = field(default_factory=list)
-    ddl: bool = True
-    sequences: bool = True
     truncate: bool = False
-    data: bool = True
-    follow_foreign_keys: bool = False
-    strict: bool = False
 
     @classmethod
-    def from_stack(cls, stack: ConfigStack, root_location: Optional[str] = None):
-        ddl = stack.get("ddl", True)
-        if isinstance(ddl, dict):
-            ddl = ddl["enabled"]
-
-        location = join_paths(root_location, stack.get("location"))
+    def from_stack(cls, stack: ConfigStack, root_location: str | None = None):
+        values = cls.collect_values(stack, root_location)
 
         return from_partial(
             cls,
-            name=stack.get("name"),
-            location=location,
+            **values,
             strategy=stack.get("strategy"),
             truncate=stack.get("truncate"),
-            compression=stack.get("compression"),
-            exclude=stack.get("exclude"),
-            sequences=stack.get("sequences", True),
-            data=stack.get("data", True),
-            ddl=stack.get("ddl", True),
-            follow_foreign_keys=stack.get("follow_foreign_keys", False),
-            strict=stack.get("strict", False),
         )
 
 
@@ -265,7 +322,7 @@ class RestoreConfig(TableParentConfig[RestoreTableConfig]):
         return RestoreTableConfig
 
 
-def normalize_table_config(tables_config: typing.Union[list, dict]) -> list:
+def normalize_table_config(tables_config: list | dict) -> list:
     """Convert the dict-style table declaration into list style.
 
     from: {"name": {"location": "backups/{table}"}}
@@ -276,9 +333,10 @@ def normalize_table_config(tables_config: typing.Union[list, dict]) -> list:
     to: [{"name": "name1"}, {"name": "name2"}]
     """
     if isinstance(tables_config, dict):
-        tables_config = [{"name": k, **v} for k, v in tables_config.items()]
-    elif isinstance(tables_config, list):
-        tables_config = [{"name": c} if isinstance(c, str) else c for c in tables_config]
+        return [{"name": k, **v} for k, v in tables_config.items()]
+
+    if isinstance(tables_config, list):
+        return [{"name": c} if isinstance(c, str) else c for c in tables_config]
 
     return tables_config
 
@@ -288,38 +346,15 @@ def normalize_table_config(tables_config: typing.Union[list, dict]) -> list:
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 
 
-@dataclass
-class SentryConfig(Config):
-    """Configuration for Sentry integration.
-
-    SentryConfig does not inherit keys from the root config.
-    """
-
-    dsn: Optional[str] = None
-    environment: Optional[str] = None
-    version: Optional[str] = None
-
-    @classmethod
-    def from_dict(cls, sentry_config: Optional[dict]) -> Optional[SentryConfig]:
-        if sentry_config is None:
-            return None
-
-        return cls(
-            dsn=sentry_config.get("dsn"),
-            environment=sentry_config.get("environment"),
-            version=sentry_config.get("version"),
-        )
-
-
 @dataclass
 class S3Config(Config):
-    aws_access_key_id: Optional[str] = None
-    aws_secret_access_key: Optional[str] = None
-    region: Optional[str] = None
-    profile: Optional[str] = None
+    aws_access_key_id: str | None = None
+    aws_secret_access_key: str | None = None
+    region: str | None = None
+    profile: str | None = None
 
     @classmethod
-    def from_dict(cls, s3_config: Optional[dict]) -> Optional[S3Config]:
+    def from_dict(cls, s3_config: dict | None) -> S3Config | None:
         if s3_config is None:
             return None
         return cls(
@@ -328,3 +363,10 @@ class S3Config(Config):
             region=s3_config.get("region"),
             profile=s3_config.get("profile"),
         )
+
+
+def compose_root_location(root_location, location, *, default):
+    if root_location is None:
+        return location or default
+
+    return join_paths(root_location, location or default)
